import logging

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from django.views.generic.base import View
from extras.signals import clear_webhooks
from netbox.views import generic
from tenancy.views import ObjectContactsView
from utilities.exceptions import AbortRequest, PermissionsViolation
from utilities.forms import restrict_form_fields
from utilities.utils import count_related, prepare_cloned_fields
from utilities.views import GetReturnURLMixin, ViewTab, register_model_view

from . import constants, exceptions, filtersets, forms, models, tables, utils

plugin_settings = settings.PLUGINS_CONFIG.get('netbox_secrets')

#
# Mixins
#


class ObjectChildrenViewMixin(generic.ObjectChildrenView):
    def get_extra_context(self, request, instance):
        return {
            'table_config': f'{self.table.__name__}_config',
        }


#
# Secret roles
#


class SecretRoleListView(generic.ObjectListView):
    queryset = models.SecretRole.objects.annotate(secret_count=count_related(models.Secret, 'role')).prefetch_related(
        'tags',
    )
    table = tables.SecretRoleTable
    filterset = filtersets.SecretRoleFilterSet
    filterset_form = forms.SecretRoleFilterForm


@register_model_view(models.SecretRole)
class SecretRoleView(generic.ObjectView):
    queryset = models.SecretRole.objects.prefetch_related('tags')


@register_model_view(models.SecretRole, 'secret')
class SecretRoleSecretView(ObjectChildrenViewMixin):
    queryset = models.SecretRole.objects.all()
    child_model = models.Secret
    table = tables.SecretTable
    filterset = filtersets.SecretFilterSet
    template_name = 'netbox_secrets/inc/view_tab.html'
    tab = ViewTab(
        label=_('Secrets'),
        badge=lambda obj: models.Secret.objects.filter(role=obj).count(),
        weight=500,
        hide_if_empty=True,
    )

    def get_children(self, request, parent):
        return models.Secret.objects.filter(role=parent)


@register_model_view(models.SecretRole, 'edit')
class SecretRoleEditView(generic.ObjectEditView):
    queryset = models.SecretRole.objects.prefetch_related('tags')
    form = forms.SecretRoleForm


@register_model_view(models.SecretRole, 'delete')
class SecretRoleDeleteView(generic.ObjectDeleteView):
    queryset = models.SecretRole.objects.prefetch_related('tags')


class SecretRoleBulkImportView(generic.BulkImportView):
    queryset = models.SecretRole.objects.prefetch_related('tags')
    model_form = forms.SecretRoleImportForm
    table = tables.SecretRoleTable


class SecretRoleBulkEditView(generic.BulkEditView):
    queryset = models.SecretRole.objects.annotate(secret_count=count_related(models.Secret, 'role')).prefetch_related(
        'tags',
    )
    filterset = filtersets.SecretRoleFilterSet
    table = tables.SecretRoleTable
    form = forms.SecretRoleBulkEditForm


class SecretRoleBulkDeleteView(generic.BulkDeleteView):
    queryset = models.SecretRole.objects.annotate(secret_count=count_related(models.Secret, 'role')).prefetch_related(
        'tags',
    )
    table = tables.SecretRoleTable


#
# Secrets
#


class SecretListView(generic.ObjectListView):
    queryset = models.Secret.objects.prefetch_related('role', 'tags')
    filterset = filtersets.SecretFilterSet
    filterset_form = forms.SecretFilterForm
    table = tables.SecretTable
    actions = ('bulk_delete', 'bulk_edit')


@register_model_view(models.Secret)
class SecretView(generic.ObjectView):
    queryset = models.Secret.objects.prefetch_related('role', 'tags')


@register_model_view(models.Secret, 'edit')
class SecretEditView(generic.ObjectEditView):
    queryset = models.Secret.objects.prefetch_related('role', 'tags')
    form = forms.SecretForm
    template_name = 'netbox_secrets/secret_edit.html'

    def alter_object(self, instance, request, url_args, url_kwargs):
        if not instance.pk:
            # Assign the assigned_object based on the URL parameters
            content_type = get_object_or_404(ContentType, pk=request.GET.get('assigned_object_type'))
            instance.assigned_object = get_object_or_404(
                content_type.model_class(),
                pk=request.GET.get('assigned_object_id'),
            )

        return instance

    def get_extra_addanother_params(self, request):
        return {
            'assigned_object_type': request.GET.get('assigned_object_type'),
            'assigned_object_id': request.GET.get('assigned_object_id'),
        }

    def dispatch(self, request, *args, **kwargs):

        # Check that the user has a valid UserKey
        try:
            uk = models.UserKey.objects.get(user=request.user)
        except models.UserKey.DoesNotExist:
            messages.warning(request, "This operation requires an active user key, but you don't have one.")
            return redirect('plugins:netbox_secrets:userkey')
        if not uk.is_active():
            messages.warning(request, "This operation is not available. Your user key has not been activated.")
            return redirect('plugins:netbox_secrets:userkey')

        return super().dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        logger = logging.getLogger('netbox.views.ObjectEditView')
        obj = self.get_object(**kwargs)
        session_key = utils.get_session_key(request)

        # Take a snapshot for change logging (if editing an existing object)
        if obj.pk and hasattr(obj, 'snapshot'):
            obj.snapshot()

        obj = self.alter_object(obj, request, args, kwargs)

        form = self.form(data=request.POST, instance=obj)
        restrict_form_fields(form, request.user)

        if form.is_valid():
            logger.debug("Form validation was successful")
            try:
                with transaction.atomic():
                    object_created = form.instance.pk is None
                    obj = form.save(commit=False)

                    # We must have a session key in order to set the plaintext of a Secret
                    if form.cleaned_data['plaintext'] and session_key is None:
                        logger.debug("Unable to proceed: No session key was provided with the request")
                        form.add_error(
                            None,
                            "No session key was provided with the request. Unable to encrypt secret data.",
                        )

                    elif form.cleaned_data['plaintext']:
                        master_key = None
                        try:
                            sk = models.SessionKey.objects.get(userkey__user=request.user)
                            master_key = sk.get_master_key(session_key)
                        except models.SessionKey.DoesNotExist:
                            logger.debug("Unable to proceed: User has no session key assigned")
                            form.add_error(None, "No session key found for this user.")
                        except exceptions.InvalidKey:
                            logger.debug("Unable to proceed: Session key is invalid")
                            form.add_error(None, "Invalid session key provided.")

                        if master_key is not None:
                            logger.debug("Successfully resolved master key for encryption")
                            obj.plaintext = str(form.cleaned_data['plaintext'])
                            obj.encrypt(master_key)

                    if form.errors:
                        logger.debug("Form validation failed")
                        return render(
                            request,
                            self.template_name,
                            {
                                'object': obj,
                                'form': form,
                                'return_url': self.get_return_url(request, obj),
                                **self.get_extra_context(request, obj),
                            },
                        )

                    obj.save()
                    form.save_m2m()

                    msg = '{} {}'.format(
                        'Created' if object_created else 'Modified',
                        self.queryset.model._meta.verbose_name,
                    )
                    logger.info(f"{msg} {obj} (PK: {obj.pk})")
                    if hasattr(obj, 'get_absolute_url'):
                        msg = mark_safe(f'{msg} <a href="{obj.get_absolute_url()}">{escape(obj)}</a>')
                    else:
                        msg = f'{msg} {obj}'
                    messages.success(request, msg)

                    if '_addanother' in request.POST:
                        redirect_url = request.path

                        # If cloning is supported, pre-populate a new instance of the form
                        params = prepare_cloned_fields(obj)
                        params.update(self.get_extra_addanother_params(request))
                        if params:
                            if 'return_url' in request.GET:
                                params['return_url'] = request.GET.get('return_url')
                            redirect_url += f"?{params.urlencode()}"

                        return redirect(redirect_url)

                    return_url = self.get_return_url(request, obj)

                    return redirect(return_url)
            except (AbortRequest, PermissionsViolation) as e:
                logger.debug(e.message)
                form.add_error(None, e.message)
                clear_webhooks.send(sender=self)

        else:
            logger.debug("Form validation failed")

        return render(
            request,
            self.template_name,
            {
                'object': obj,
                'form': form,
                'return_url': self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )


@register_model_view(models.Secret, 'delete')
class SecretDeleteView(generic.ObjectDeleteView):
    queryset = models.Secret.objects.prefetch_related('role', 'tags')


class SecretBulkEditView(generic.BulkEditView):
    queryset = models.Secret.objects.prefetch_related('role', 'tags')
    filterset = filtersets.SecretFilterSet
    table = tables.SecretTable
    form = forms.SecretBulkEditForm


class SecretBulkDeleteView(generic.BulkDeleteView):
    queryset = models.Secret.objects.prefetch_related('role', 'tags')
    filterset = filtersets.SecretFilterSet
    table = tables.SecretTable


if plugin_settings.get('enable_contacts'):
    @register_model_view(models.Secret, 'contacts')
    class SecretContactsView(ObjectContactsView):
        queryset = models.Secret.objects.prefetch_related('role', 'tags')


class UserKeyView(LoginRequiredMixin, View):
    template_name = 'netbox_secrets/userkey.html'

    def get(self, request):
        try:
            userkey = models.UserKey.objects.get(user=request.user)
        except models.UserKey.DoesNotExist:
            userkey = None

        return render(
            request,
            self.template_name,
            {
                'object': userkey,
            },
        )


class UserKeyEditView(LoginRequiredMixin, GetReturnURLMixin, View):
    queryset = models.SessionKey.objects.all()
    template_name = 'netbox_secrets/userkey_edit.html'

    def dispatch(self, request, *args, **kwargs):
        try:
            self.userkey = models.UserKey.objects.get(user=request.user)
        except models.UserKey.DoesNotExist:
            self.userkey = models.UserKey(user=request.user)

        return super().dispatch(request, *args, **kwargs)

    def get(self, request):
        form = forms.UserKeyForm(instance=self.userkey)
        return render(
            request,
            self.template_name,
            {
                'object': self.userkey,
                'form': form,
                'return_url': self.get_return_url(request, self.userkey),
            },
        )

    def post(self, request):
        logger = logging.getLogger('netbox.views.ObjectEditView')
        form = forms.UserKeyForm(data=request.POST, instance=self.userkey)
        if form.is_valid():
            uk = form.save(commit=False)
            uk.user = request.user
            uk.save()
            messages.success(request, "Your user key has been saved.")
            return redirect('plugins:netbox_secrets:userkey')
        else:
            logger.debug("Form validation failed")
            messages.error(request, "Unable to save your user key.")

        return render(
            request,
            self.template_name,
            {
                'userkey': self.userkey,
                'form': form,
            },
        )


@register_model_view(models.SessionKey, 'delete')
class SessionKeyDeleteView(generic.ObjectDeleteView):
    queryset = models.SessionKey.objects.all()

    def get_queryset(self, request):
        return super().get_queryset(request).filter(userkey__user=request.user)

    def post(self, request, *args, **kwargs):
        response = super().post(request, *args, **kwargs)
        # Delete the cookie
        response.delete_cookie(constants.SESSION_COOKIE_NAME)
        return response



from netbox.forms import NetBoxModelImportForm
from utilities.forms.fields import SlugField

from ..models import SecretRole

__all__ = [
    'SecretRoleImportForm',
]


class SecretRoleImportForm(NetBoxModelImportForm):
    slug = SlugField()

    class Meta:
        model = SecretRole
        fields = ('name', 'slug')



from django.conf import settings
from django.db.models import Q

plugin_settings = settings.PLUGINS_CONFIG.get('netbox_secrets')

#
# Secrets
#

SECRET_ASSIGNABLE_MODELS = Q()
for app_model in plugin_settings.get('apps'):
    app_label, model = app_model.split('.')
    SECRET_ASSIGNABLE_MODELS |= Q(app_label=app_label, model=model)

SECRET_PLAINTEXT_MAX_LENGTH = 65535


#
# Session Keys
#

SESSION_COOKIE_NAME = 'netbox_secrets_sessionid'



# Generated by Django 4.1.9 on 2023-05-07 18:36

from django.contrib.contenttypes.models import ContentType
from django.db import migrations, models


def copy_assigned_object(apps, schema_editor):
    Secret = apps.get_model('netbox_secrets', 'Secret')
    try:
        for secret in Secret.objects.all():
            content_type = ContentType.objects.get(id=secret.assigned_object_type_id)
            Model = apps.get_model(content_type.app_label, str(content_type.model).capitalize())
            object = Model.objects.filter(id=secret.assigned_object_id).first()
            secret._object_repr = object.name
            secret.save()
    except:
        pass


class Migration(migrations.Migration):
    dependencies = [
        ('netbox_secrets', '0006_alter_userkey_created_alter_userkey_last_updated'),
    ]

    operations = [
        migrations.AddField(
            model_name='secret',
            name='_object_repr',
            field=models.CharField(blank=True, editable=False, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name='secret',
            name='comments',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='secret',
            name='description',
            field=models.CharField(blank=True, max_length=200),
        ),
        migrations.AddField(
            model_name='secretrole',
            name='comments',
            field=models.TextField(blank=True),
        ),
        migrations.RunPython(copy_assigned_object)
    ]



import string

from Crypto.PublicKey import RSA
from django.conf import settings
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.test import TestCase

from netbox_secrets.models import Secret, UserKey
from netbox_secrets.utils.crypto import (
    decrypt_master_key,
    encrypt_master_key,
    generate_random_key,
)
from netbox_secrets.utils.hashers import SecretValidationHasher


class UserKeyTestCase(TestCase):
    def setUp(self):
        self.TEST_KEYS = {}
        key_size = settings.PLUGINS_CONFIG['netbox_secrets'].get('public_key_size')
        for username in ['alice', 'bob']:
            User.objects.create_user(username=username, password=username)
            key = RSA.generate(key_size)
            self.TEST_KEYS[f'{username}_public'] = key.publickey().exportKey('PEM')
            self.TEST_KEYS[f'{username}_private'] = key.exportKey('PEM')

    def test_01_fill(self):
        """
        Validate the filling of a UserKey with public key material.
        """
        alice_uk = UserKey(user=User.objects.get(username='alice'))
        self.assertFalse(alice_uk.is_filled(), "UserKey with empty public_key is_filled() did not return False")
        alice_uk.public_key = self.TEST_KEYS['alice_public']
        self.assertTrue(alice_uk.is_filled(), "UserKey with public key is_filled() did not return True")

    def test_02_activate(self):
        """
        Validate the activation of a UserKey.
        """
        master_key = generate_random_key()
        alice_uk = UserKey(user=User.objects.get(username='alice'), public_key=self.TEST_KEYS['alice_public'])
        self.assertFalse(alice_uk.is_active(), "Inactive UserKey is_active() did not return False")
        alice_uk.activate(master_key)
        self.assertTrue(alice_uk.is_active(), "ActiveUserKey is_active() did not return True")

    def test_03_key_sizes(self):
        """
        Ensure that RSA keys which are too small or too large are rejected.
        """
        rsa = RSA.generate(settings.PLUGINS_CONFIG['netbox_secrets'].get('public_key_size', 2048) - 256)
        small_key = rsa.publickey().exportKey('PEM')
        try:
            UserKey(public_key=small_key).clean()
            self.fail("UserKey.clean() did not fail with an undersized RSA key")
        except ValidationError:
            pass
        rsa = RSA.generate(4096 + 256)  # Max size is 4096 (enforced by master_key_cipher field size)
        big_key = rsa.publickey().exportKey('PEM')
        try:
            UserKey(public_key=big_key).clean()
            self.fail("UserKey.clean() did not fail with an oversized RSA key")
        except ValidationError:
            pass

    def test_04_master_key_retrieval(self):
        """
        Test the decryption of a master key using the user's private key.
        """
        master_key = generate_random_key()
        alice_uk = UserKey(user=User.objects.get(username='alice'), public_key=self.TEST_KEYS['alice_public'])
        alice_uk.activate(master_key)
        retrieved_master_key = alice_uk.get_master_key(self.TEST_KEYS['alice_private'])
        self.assertEqual(master_key, retrieved_master_key, "Master key retrieval failed with correct private key")

    def test_05_invalid_private_key(self):
        """
        Ensure that an exception is raised when attempting to retrieve a secret key using an invalid private key.
        """
        secret_key = generate_random_key()
        secret_key_cipher = encrypt_master_key(secret_key, self.TEST_KEYS['alice_public'])
        try:
            decrypt_master_key(secret_key_cipher, self.TEST_KEYS['bob_private'])
            self.fail("Decrypting secret key from Alice's UserKey using Bob's private key did not fail")
        except ValueError:
            pass


class SecretTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Generate a random key for encryption/decryption of secrets
        cls.secret_key = generate_random_key()

    def test_01_encrypt_decrypt(self):
        """
        Test basic encryption and decryption functionality using a random master key.
        """
        plaintext = string.printable * 2
        s = Secret(plaintext=plaintext)
        s.encrypt(self.secret_key)

        # Ensure plaintext is deleted upon encryption
        self.assertIsNone(s.plaintext, "Plaintext must be None after encrypting.")

        # Enforce minimum ciphertext length
        self.assertGreaterEqual(len(s.ciphertext), 80, "Ciphertext must be at least 80 bytes (16B IV + 64B+ ciphertext")

        # Ensure proper hashing algorithm is used
        hasher, iterations, salt, sha256 = s.hash.split('$')
        self.assertEqual(hasher, 'pbkdf2_sha256', f"Hashing algorithm has been modified to: {hasher}")
        self.assertGreaterEqual(
            int(iterations),
            SecretValidationHasher.iterations,
            f"Insufficient iteration count ({iterations}) for hash",
        )
        self.assertGreaterEqual(len(salt), 12, f"Hash salt is too short ({len(salt)} chars)")

        # Test hash validation
        self.assertTrue(s.validate(plaintext), "Plaintext does not validate against the generated hash")
        self.assertFalse(s.validate(""), "Empty plaintext validated against hash")
        self.assertFalse(s.validate("Invalid plaintext"), "Invalid plaintext validated against hash")

        # Test decryption
        s.decrypt(self.secret_key)
        self.assertEqual(plaintext, s.plaintext, "Decrypting Secret returned incorrect plaintext")

    def test_02_ciphertext_uniqueness(self):
        """
        Generate 50 Secrets using the same plaintext and check for duplicate IVs or payloads.
        """
        plaintext = "1234567890abcdef"
        ivs = []
        ciphertexts = []
        for i in range(1, 51):
            s = Secret(plaintext=plaintext)
            s.encrypt(self.secret_key)
            ivs.append(s.ciphertext[0:16])
            ciphertexts.append(s.ciphertext[16:32])
        duplicate_ivs = [i for i, x in enumerate(ivs) if ivs.count(x) > 1]
        self.assertEqual(duplicate_ivs, [], "One or more duplicate IVs found!")
        duplicate_ciphertexts = [i for i, x in enumerate(ciphertexts) if ciphertexts.count(x) > 1]
        self.assertEqual(duplicate_ciphertexts, [], "One or more duplicate ciphertexts (first blocks) found!")

    def test_minimum_length(self):
        """
        Test enforcement of the minimum length for ciphertexts.
        """
        plaintext = 'A'  # One-byte plaintext
        secret = Secret(plaintext=plaintext)
        secret.encrypt(self.secret_key)

        # 16B IV + 2B length + 1B secret + 61B padding = 80 bytes
        self.assertEqual(len(secret.ciphertext), 80)
        self.assertIsNone(secret.plaintext)

        secret.decrypt(self.secret_key)
        self.assertEqual(secret.plaintext, plaintext)

    def test_maximum_length(self):
        """
        Test encrypting a plaintext value of the maximum length.
        """
        plaintext = '0123456789abcdef' * 4096
        plaintext = plaintext[:65535]  # 65,535 chars
        secret = Secret(plaintext=plaintext)
        secret.encrypt(self.secret_key)

        # 16B IV + 2B length + 65535B secret + 15B padding = 65568 bytes
        self.assertEqual(len(secret.ciphertext), 65568)
        self.assertIsNone(secret.plaintext)

        secret.decrypt(self.secret_key)
        self.assertEqual(secret.plaintext, plaintext)



from django import forms
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import gettext as _

from netbox.forms import NetBoxModelFilterSetForm
from utilities.forms.fields import (
    ContentTypeMultipleChoiceField,
    DynamicModelMultipleChoiceField,
    TagFilterField,
)
from ..constants import *
from ..models import Secret, SecretRole

__all__ = [
    'SecretRoleFilterForm',
    'SecretFilterForm',
]


class SecretRoleFilterForm(NetBoxModelFilterSetForm):
    model = SecretRole
    q = forms.CharField(required=False, label=_('Search'))
    id = DynamicModelMultipleChoiceField(queryset=SecretRole.objects.all(), required=False, label=_('Roles'))


class SecretFilterForm(NetBoxModelFilterSetForm):
    model = Secret

    fieldsets = (
        (None, ('q', 'filter_id', 'tag')),
        ('Attributes', ('role_id', 'assigned_object_type_id')),
    )

    q = forms.CharField(required=False, label=_('Search'))
    assigned_object_type_id = ContentTypeMultipleChoiceField(
        queryset=ContentType.objects.filter(SECRET_ASSIGNABLE_MODELS),
        required=False,
        label='Object type(s)',
    )
    role_id = DynamicModelMultipleChoiceField(queryset=SecretRole.objects.all(), required=False, label=_('Role'))
    tag = TagFilterField(model)



from django.db import migrations


def populate_secrets(apps, schema_editor):
    """Populate the Secret model with data from the SecretStore model."""
    try:
        SecretOld = apps.get_model('netbox_secretstore', 'Secret')
    except LookupError:
        # Skip if the old model doesn't exist
        return
    Secret = apps.get_model('netbox_secrets', 'Secret')

    # Retrieve the necessary data from SecretStore objects
    secrets = SecretOld.objects.values(
        'created',
        'last_updated',
        'custom_field_data',
        'id',
        'assigned_object_id',
        'name',
        'ciphertext',
        'hash',
        'assigned_object_type_id',
        'role_id',
    )

    # Queue Secrets to be created
    secrets_to_create = []
    for i, secret in enumerate(secrets, start=1):
        secrets_to_create.append(
            Secret(
                created=secret['created'],
                last_updated=secret['last_updated'],
                custom_field_data=secret['custom_field_data'],
                id=secret['id'],
                assigned_object_id=secret['assigned_object_id'],
                name=secret['name'],
                ciphertext=secret['ciphertext'],
                assigned_object_type_id=secret['assigned_object_type_id'],
                role_id=secret['role_id'],
                hash=secret['hash'],
            ),
        )

    # Bulk create the secret objects
    Secret.objects.bulk_create(secrets_to_create, batch_size=100)


def update_objectchanges(apps, schema_editor):
    """
    Update the ObjectChange records to reflect the new model name.
    """
    ObjectChange = apps.get_model('extras', 'ObjectChange')
    ContentType = apps.get_model('contenttypes', 'ContentType')

    try:
        ctsecret = ContentType.objects.get_for_model(apps.get_model('netbox_secrets', 'Secret'))
        ctsecretrole = ContentType.objects.get_for_model(apps.get_model('netbox_secrets', 'SecretRole'))
        ctuserkey = ContentType.objects.get_for_model(apps.get_model('netbox_secrets', 'UserKey'))
        ctsessionkey = ContentType.objects.get_for_model(apps.get_model('netbox_secrets', 'SessionKey'))

        ctnbsecret = ContentType.objects.get(app_label='netbox_secretstore', model='secret')
        ctnbsecretrole = ContentType.objects.get(app_label='netbox_secretstore', model='secretrole')
        ctnbuserkey = ContentType.objects.get(app_label='netbox_secretstore', model='userkey')
        ctnbsessionkey = ContentType.objects.get(app_label='netbox_secretstore', model='sessionkey')

        ObjectChange.objects.filter(changed_object_type_id=ctnbsecret.id).update(changed_object_type_id=ctsecret.id)
        ObjectChange.objects.filter(changed_object_type_id=ctnbsecretrole.id).update(
            changed_object_type_id=ctsecretrole.id,
        )
        ObjectChange.objects.filter(changed_object_type_id=ctnbsessionkey.id).update(
            changed_object_type_id=ctsessionkey.id,
        )
        ObjectChange.objects.filter(changed_object_type_id=ctnbuserkey.id).update(changed_object_type_id=ctuserkey.id)
    except (ContentType.DoesNotExist, LookupError):
        pass


class Migration(migrations.Migration):
    dependencies = [
        ('netbox_secrets', '0003_populate_secretroles'),
    ]

    operations = [
        migrations.RunPython(code=populate_secrets, reverse_code=migrations.RunPython.noop),
        migrations.RunPython(code=update_objectchanges, reverse_code=migrations.RunPython.noop),
    ]



from django.db import migrations


def populate_userkeys(apps, schema_editor):
    """Populate the UserKey model with data from the SecretStore model."""
    try:
        UserKeyOld = apps.get_model('netbox_secretstore', 'UserKey')
    except LookupError:
        # Skip if the old model doesn't exist
        return
    UserKey = apps.get_model('netbox_secrets', 'UserKey')

    # Retrieve the necessary data from SecretStore objects
    userkeys = UserKeyOld.objects.values('id', 'created', 'last_updated', 'user', 'public_key', 'master_key_cipher')

    # Queue UserKeys to be created
    userkeys_to_create = []
    for i, userkey in enumerate(userkeys, start=1):
        userkeys_to_create.append(
            UserKey(
                id=userkey['id'],
                created=userkey['created'],
                last_updated=userkey['last_updated'],
                user_id=userkey['user'],
                public_key=userkey['public_key'],
                master_key_cipher=userkey['master_key_cipher'],
            ),
        )

    # Bulk create the userkey objects
    UserKey.objects.bulk_create(userkeys_to_create, batch_size=100)


class Migration(migrations.Migration):
    dependencies = [
        ('netbox_secrets', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(code=populate_userkeys, reverse_code=migrations.RunPython.noop),
    ]



from netbox.graphql.types import NetBoxObjectType, ObjectType

from netbox_secrets import filtersets, models

__all__ = [
    'SecretRoleType',
    'SecretType',
]


class SecretRoleType(ObjectType):
    class Meta:
        model = models.SecretRole
        fields = '__all__'
        filterset_class = filtersets.SecretRoleFilterSet


class SecretType(NetBoxObjectType):
    class Meta:
        model = models.Secret
        fields = [
            'id',
            'name',
            'description',
            'role',
            'assigned_object_type',
            'assigned_object_id',
            'comments',
            'tags',
            'created',
            'last_updated',
        ]
        filterset_class = filtersets.SecretFilterSet



import graphene
from netbox.graphql.fields import ObjectField, ObjectListField

from .types import *


class SecretsQuery(graphene.ObjectType):
    secret = ObjectField(SecretType)
    secret_list = ObjectListField(SecretType)

    secretrole = ObjectField(SecretRoleType)
    secretrole_list = ObjectListField(SecretRoleType)



import logging

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.utils import OperationalError
from extras.plugins import PluginTemplateExtension
from netbox.views import generic
from utilities.views import ViewTab, register_model_view

from .filtersets import SecretFilterSet
from .models import Secret
from .tables import SecretTable

logger = logging.getLogger(__name__)
plugin_settings = settings.PLUGINS_CONFIG.get('netbox_secrets')
template_extensions = []


def secrets_panel(self):
    obj = self.context['object']
    app_label, model = self.model.split('.')
    assigned_object_type = ContentType.objects.get(app_label=app_label, model=model).id

    return self.render(
        'netbox_secrets/inc/secrets_panel.html',
        extra_context={
            'secrets': Secret.objects.filter(assigned_object_type=assigned_object_type, assigned_object_id=obj.id),
        },
    )


def get_display_on(app_model):
    """Get preferred display location for app_model"""
    display_on = plugin_settings.get('display_default')

    if display_setting := plugin_settings.get('display_setting'):
        display_on = display_setting.get(app_model, display_on)

    return display_on


def tab_view(_model):
    class ModelTabView(generic.ObjectChildrenView):
        queryset = _model.objects.all()
        child_model = Secret
        table = SecretTable
        filterset = SecretFilterSet
        template_name = 'netbox_secrets/inc/view_tab.html'
        tab = ViewTab(
            label='Secrets',
            badge=lambda obj: obj.secrets.count(),
            weight=500,
        )

        def get_children(self, request, parent):
            return self.child_model.objects.restrict(request.user, 'view').filter(
                assigned_object_type=ContentType.objects.get_for_model(parent),
                assigned_object_id=parent.pk,
            )

    register_model_view(_model, name='secrets')(ModelTabView)


def secret_add_button(_app_model):
    class Button(PluginTemplateExtension):
        model = _app_model

        def buttons(self):
            return self.render(
                'netbox_secrets/inc/secret_add_button.html',
            )

    return Button


# Generate plugin extensions for the defined classes
try:
    for app_model in plugin_settings.get('apps'):
        app_label, model = app_model.split('.')
        klass_name = f'{app_label}_{model}_plugin_template_extension'

        display = get_display_on(app_model)

        if display == 'tab_view':
            template_extensions.append(secret_add_button(app_model))
            tab_view(ContentType.objects.get(app_label=app_label, model=model).model_class())
        else:
            dynamic_klass = type(
                klass_name,
                (PluginTemplateExtension,),
                {'model': app_model, get_display_on(app_model): secrets_panel},
            )
            template_extensions.append(dynamic_klass)
except OperationalError as e:
    # This happens when the database is not yet ready
    logger.warning(f'Database not ready, skipping plugin extensions: {e}')
except Exception as e:
    # Unexpected error
    raise Exception(f'Unexpected error: {e}')



# Generated by Django 4.0.8 on 2022-12-28 11:35

import django.core.serializers.json
import django.db.models.deletion
import taggit.managers
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('extras', '0077_customlink_extend_text_and_url'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='UserKey',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('created', models.DateField(auto_now_add=True)),
                ('last_updated', models.DateTimeField(auto_now=True)),
                ('public_key', models.TextField()),
                ('master_key_cipher', models.BinaryField(blank=True, max_length=512, null=True)),
                (
                    'user',
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='user_key',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'ordering': ['user__username'],
            },
        ),
        migrations.CreateModel(
            name='SessionKey',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('cipher', models.BinaryField(max_length=512)),
                ('hash', models.CharField(editable=False, max_length=128)),
                ('created', models.DateTimeField(auto_now_add=True)),
                (
                    'userkey',
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='session_key',
                        to='netbox_secrets.userkey',
                    ),
                ),
            ],
            options={
                'ordering': ['userkey__user__username'],
            },
        ),
        migrations.CreateModel(
            name='SecretRole',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                (
                    'custom_field_data',
                    models.JSONField(blank=True, default=dict, encoder=django.core.serializers.json.DjangoJSONEncoder),
                ),
                ('name', models.CharField(max_length=100, unique=True)),
                ('slug', models.SlugField(max_length=100, unique=True)),
                ('description', models.CharField(blank=True, max_length=200)),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Secret',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                (
                    'custom_field_data',
                    models.JSONField(blank=True, default=dict, encoder=django.core.serializers.json.DjangoJSONEncoder),
                ),
                ('assigned_object_id', models.PositiveIntegerField()),
                ('name', models.CharField(blank=True, max_length=100)),
                ('ciphertext', models.BinaryField(max_length=65568)),
                ('hash', models.CharField(editable=False, max_length=128)),
                (
                    'assigned_object_type',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name='secrets',
                        to='contenttypes.contenttype',
                    ),
                ),
                (
                    'role',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name='secrets',
                        to='netbox_secrets.secretrole',
                    ),
                ),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'ordering': ('role', 'name', 'pk'),
                'unique_together': {('assigned_object_type', 'assigned_object_id', 'role', 'name')},
            },
        ),
    ]



from importlib.metadata import metadata

from django.db.utils import OperationalError, ProgrammingError
from extras.plugins import PluginConfig

metadata = metadata('netbox_secrets')


class NetBoxSecrets(PluginConfig):
    name = metadata.get('Name').replace('-', '_')
    verbose_name = metadata.get('Summary')
    description = metadata.get('Description')
    version = metadata.get('Version')
    author = metadata.get('Author')
    author_email = metadata.get('Author-email')
    base_url = 'secrets'
    min_version = '3.5.0'
    max_version = '3.5.99'
    required_settings = []
    default_settings = {
        'apps': ['dcim.device', 'virtualization.virtualmachine'],
        'display_default': 'left_page',
        'display_setting': {},
        'enable_contacts': False,
        'public_key_size': 2048,
    }

    def ready(self):
        super().ready()

        # Configure generic relations for secrets
        from django.contrib.contenttypes.fields import GenericRelation
        from django.contrib.contenttypes.models import ContentType

        from .constants import SECRET_ASSIGNABLE_MODELS
        from .models import Secret

        try:
            for content_type in ContentType.objects.filter(SECRET_ASSIGNABLE_MODELS):
                GenericRelation(
                    to=Secret,
                    content_type_field='assigned_object_type',
                    object_id_field='assigned_object_id',
                    related_query_name=content_type.model,
                ).contribute_to_class(
                    content_type.model_class(),
                    'secrets',
                )
        except (OperationalError, ProgrammingError):
            pass


config = NetBoxSecrets



from extras.plugins import PluginMenuButton, PluginMenuItem
from utilities.choices import ButtonColorChoices

menu_items = (
    PluginMenuItem(
        link_text="User Key",
        link="plugins:netbox_secrets:userkey",
        permissions=["netbox_secrets.view_userkey"],
    ),
    PluginMenuItem(
        link_text="Secret Roles",
        link="plugins:netbox_secrets:secretrole_list",
        permissions=["netbox_secrets.view_secretrole"],
        buttons=(
            PluginMenuButton(
                link="plugins:netbox_secrets:secretrole_add",
                title="Add Secret Role",
                icon_class="mdi mdi-plus-thick",
                color=ButtonColorChoices.GREEN,
                permissions=["netbox_secrets.add_secretrole"],
            ),
            PluginMenuButton(
                link="plugins:netbox_secrets:secretrole_import",
                title="Import Secret Role",
                icon_class="mdi mdi-upload",
                color=ButtonColorChoices.CYAN,
                permissions=["netbox_secrets.add_secretrole"],
            ),
        ),
    ),
    PluginMenuItem(
        link_text="Secrets",
        link="plugins:netbox_secrets:secret_list",
        permissions=["netbox_secrets.view_secret"],
    ),
)



from Crypto.Cipher import PKCS1_OAEP
from Crypto.PublicKey import RSA
from django import forms
from netbox.forms import NetBoxModelForm
from utilities.forms.fields import CommentField, DynamicModelChoiceField, SlugField

from ..constants import *
from ..models import Secret, SecretRole, UserKey

__all__ = [
    'ActivateUserKeyForm',
    'SecretRoleForm',
    'SecretForm',
    'UserKeyForm',
]


def validate_rsa_key(key, is_secret=True):
    """
    Validate the format and type of an RSA key.
    """
    if key.startswith('ssh-rsa '):
        raise forms.ValidationError(
            "OpenSSH line format is not supported. Please ensure that your public is in PEM (base64) format.",
        )
    try:
        key = RSA.importKey(key)
    except ValueError:
        raise forms.ValidationError("Invalid RSA key. Please ensure that your key is in PEM (base64) format.")
    except Exception as e:
        raise forms.ValidationError(f"Invalid key detected: {e}")
    if is_secret and not key.has_private():
        raise forms.ValidationError("This looks like a public key. Please provide your private RSA key.")
    elif not is_secret and key.has_private():
        raise forms.ValidationError("This looks like a private key. Please provide your public RSA key.")
    try:
        PKCS1_OAEP.new(key)
    except Exception:
        raise forms.ValidationError("Error validating RSA key. Please ensure that your key supports PKCS#1 OAEP.")


class SecretRoleForm(NetBoxModelForm):
    slug = SlugField()

    fieldsets = ((None, ('name', 'slug', 'description', 'tags')),)

    class Meta:
        model = SecretRole
        fields = ('name', 'slug', 'description', 'comments', 'tags')


class SecretForm(NetBoxModelForm):
    plaintext = forms.CharField(
        max_length=SECRET_PLAINTEXT_MAX_LENGTH,
        required=False,
        label='Plaintext',
        widget=forms.PasswordInput(
            attrs={
                'class': 'requires-session-key',
                'autocomplete': 'new-password',
            },
        ),
    )
    plaintext2 = forms.CharField(
        max_length=SECRET_PLAINTEXT_MAX_LENGTH,
        required=False,
        label='Plaintext (verify)',
        widget=forms.PasswordInput(
            attrs={
                'autocomplete': 'new-password',
            },
        ),
    )
    role = DynamicModelChoiceField(queryset=SecretRole.objects.all())

    comments = CommentField()

    fieldsets = (
        (None, ('name', 'description', 'role', 'tags')),
        ('Secret Data', ('plaintext', 'plaintext2')),
    )

    class Meta:
        model = Secret
        fields = (
            'role',
            'name',
            'plaintext',
            'plaintext2',
            'tags',
            'description',
            'comments',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # A plaintext value is required when creating a new Secret
        if not self.instance.pk:
            self.fields['plaintext'].required = True

    def clean(self):
        super().clean()

        # Verify that the provided plaintext values match
        if self.cleaned_data['plaintext'] != self.cleaned_data['plaintext2']:
            raise forms.ValidationError(
                {'plaintext2': "The two given plaintext values do not match. Please check your input."},
            )


class UserKeyForm(forms.ModelForm):
    public_key = forms.CharField(
        widget=forms.Textarea(
            attrs={
                'class': 'form-control',
            },
        ),
        label='Public Key (PEM format)',
        help_text='Enter your public RSA key. Keep the private one with you; you will need it for decryption. Please '
        'note that passphrase-protected keys are not supported.',
    )

    class Meta:
        model = UserKey
        fields = ['public_key']

    def clean_public_key(self):
        key = self.cleaned_data['public_key']

        # Validate the RSA key format.
        validate_rsa_key(key, is_secret=False)

        return key


class ActivateUserKeyForm(forms.Form):
    _selected_action = forms.ModelMultipleChoiceField(queryset=UserKey.objects.all(), label='User Keys')
    secret_key = forms.CharField(
        widget=forms.Textarea(
            attrs={
                'class': 'vLargeTextField',
            },
        ),
        label='Your private key',
    )



from django.urls import include, path
from utilities.urls import get_model_urls

from . import views

urlpatterns = [
    # Secret roles
    path('secret-roles/', views.SecretRoleListView.as_view(), name='secretrole_list'),
    path('secret-roles/add/', views.SecretRoleEditView.as_view(), name='secretrole_add'),
    path('secret-roles/import/', views.SecretRoleBulkImportView.as_view(), name='secretrole_import'),
    path('secret-roles/edit/', views.SecretRoleBulkEditView.as_view(), name='secretrole_bulk_edit'),
    path('secret-roles/delete/', views.SecretRoleBulkDeleteView.as_view(), name='secretrole_bulk_delete'),
    path('secret-roles/<int:pk>/', include(get_model_urls('netbox_secrets', 'secretrole'))),
    # Secrets
    path('secrets/', views.SecretListView.as_view(), name='secret_list'),
    path('secrets/add/', views.SecretEditView.as_view(), name='secret_add'),
    path('secrets/delete/', views.SecretBulkDeleteView.as_view(), name='secret_bulk_delete'),
    path('secrets/edit/', views.SecretBulkEditView.as_view(), name='secret_bulk_edit'),
    path('secrets/<int:pk>/', include(get_model_urls('netbox_secrets', 'secret'))),
    # User
    path('user-key/', views.UserKeyView.as_view(), name='userkey'),
    path('user-key/edit/', views.UserKeyEditView.as_view(), name='userkey_edit'),
    path('session-key/delete/', views.SessionKeyDeleteView.as_view(), name='sessionkey_delete'),
]



import base64

from Crypto.PublicKey import RSA
from django.conf import settings
from django.http import HttpResponseBadRequest
from drf_spectacular import utils as drf_utils
from rest_framework import mixins as drf_mixins
from rest_framework.exceptions import ValidationError
from rest_framework.parsers import FormParser, JSONParser, MultiPartParser
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.routers import APIRootView
from rest_framework.viewsets import ModelViewSet, ViewSet

from netbox.api.viewsets import BaseViewSet, mixins, NetBoxModelViewSet
from utilities.utils import count_related
from . import serializers
from .. import constants, exceptions, filtersets, models

plugin_settings = settings.PLUGINS_CONFIG.get('netbox_secrets', {})
public_key_size = plugin_settings.get('public_key_size')

ERR_USERKEY_MISSING = "No UserKey found for the current user."
ERR_USERKEY_INACTIVE = "UserKey has not been activated for decryption."
ERR_PRIVKEY_MISSING = "Private key was not provided."
ERR_PRIVKEY_INVALID = "Invalid private key."


class SecretsRootView(APIRootView):
    """
    Secrets API root view
    """

    def get_view_name(self):
        return 'Secrets'


#
# User Key
#
class UserKeyViewSet(ModelViewSet):
    queryset = models.UserKey.objects.all()
    serializer_class = serializers.UserKeySerializer

    def get_queryset(self):
        return super().get_queryset().filter(user=self.request.user)


#
# Secret Roles
#


class SecretRoleViewSet(NetBoxModelViewSet):
    queryset = models.SecretRole.objects.annotate(secret_count=count_related(models.Secret, 'role')).prefetch_related(
        'tags',
    )
    serializer_class = serializers.SecretRoleSerializer
    filterset_class = filtersets.SecretRoleFilterSet


#
# Secrets
#


class SecretViewSet(NetBoxModelViewSet):
    queryset = models.Secret.objects.prefetch_related('role', 'tags')
    serializer_class = serializers.SecretSerializer
    filterset_class = filtersets.SecretFilterSet

    master_key = None

    def get_serializer_context(self):

        # Make the master key available to the serializer for encrypting plaintext values
        context = super().get_serializer_context()
        context['master_key'] = self.master_key

        return context

    def initial(self, request, *args, **kwargs):

        super().initial(request, *args, **kwargs)

        if request.user.is_authenticated:

            # Read session key from HTTP cookie or header if it has been provided. The session key must be provided in
            # order to encrypt/decrypt secrets.
            if constants.SESSION_COOKIE_NAME in request.COOKIES:
                session_key = base64.b64decode(request.COOKIES[constants.SESSION_COOKIE_NAME])
            elif 'HTTP_X_SESSION_KEY' in request.META:
                session_key = base64.b64decode(request.META['HTTP_X_SESSION_KEY'])
            else:
                session_key = None

            # We can't encrypt secret plaintext without a session key.
            if self.action in ['create', 'update'] and session_key is None:
                raise ValidationError("A session key must be provided when creating or updating secrets.")

            # Attempt to retrieve the master key for encryption/decryption if a session key has been provided.
            if session_key is not None:
                try:
                    sk = models.SessionKey.objects.get(userkey__user=request.user)
                    self.master_key = sk.get_master_key(session_key)
                except (models.SessionKey.DoesNotExist, exceptions.InvalidKey):
                    raise ValidationError("Invalid session key.")

    def retrieve(self, request, *args, **kwargs):

        secret = self.get_object()

        # Attempt to decrypt the secret if the master key is known
        if self.master_key is not None:
            secret.decrypt(self.master_key)

        serializer = self.get_serializer(secret)
        return Response(serializer.data)

    def list(self, request, *args, **kwargs):

        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:

            # Attempt to decrypt all secrets if the master key is known
            if self.master_key is not None:
                secrets = []
                for secret in page:
                    secret.decrypt(self.master_key)
                    secrets.append(secret)
                serializer = self.get_serializer(secrets, many=True)
            else:
                serializer = self.get_serializer(page, many=True)

            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)


#
# Session Keys
#
class SessionKeyViewSet(
    drf_mixins.ListModelMixin,
    drf_mixins.RetrieveModelMixin,
    drf_mixins.DestroyModelMixin,
    mixins.BriefModeMixin,
    mixins.BulkDestroyModelMixin,
    mixins.ObjectValidationMixin,
    BaseViewSet,
):
    queryset = models.SessionKey.objects.prefetch_related('userkey__user')
    serializer_class = serializers.SessionKeySerializer

    def get_queryset(self):
        if self.request.user.is_authenticated:
            # Overrides self.queryset to always return the restricted key filtered by the request.user
            self.queryset = super().get_queryset().filter(userkey__user=self.request.user)
            return self.queryset

        return super().get_queryset()

    @drf_utils.extend_schema(
        request=serializers.SessionKeyCreateSerializer,
        responses={
            201: drf_utils.OpenApiResponse(
                description="Session key created successfully.",
                response=serializers.SessionKeySerializer,
            ),
            400: drf_utils.OpenApiResponse(
                description="Session key creation failed.",
                response={
                    'type': 'string',
                },
                examples=[
                    drf_utils.OpenApiExample(name=ERR_PRIVKEY_MISSING, value=ERR_PRIVKEY_MISSING),
                    drf_utils.OpenApiExample(name=ERR_USERKEY_MISSING, value=ERR_USERKEY_MISSING),
                    drf_utils.OpenApiExample(name=ERR_USERKEY_INACTIVE, value=ERR_USERKEY_INACTIVE),
                    drf_utils.OpenApiExample(name=ERR_PRIVKEY_INVALID, value=ERR_PRIVKEY_INVALID),
                ],
            ),
        },
    )
    def create(self, request):
        """
        Creates a new session key for the current user.
        """

        private_key = request.data.get('private_key', None)
        preserve_key = str(request.data.get('preserve_key', False)).lower() in ['true', 'yes', '1']

        if private_key is None:
            return HttpResponseBadRequest(ERR_PRIVKEY_MISSING)

        # Validate user key
        try:
            user_key = models.UserKey.objects.get(user=request.user)
        except models.UserKey.DoesNotExist:
            return HttpResponseBadRequest(ERR_USERKEY_MISSING)
        if not user_key.is_active():
            return HttpResponseBadRequest(ERR_USERKEY_INACTIVE)

        # Validate private key
        master_key = user_key.get_master_key(private_key)
        if master_key is None:
            return HttpResponseBadRequest(ERR_PRIVKEY_INVALID)

        current_session_key = self.queryset.first()

        if current_session_key and preserve_key:

            # Retrieve the existing session key
            key = current_session_key.get_session_key(master_key)
            self.queryset = current_session_key

        else:

            # Create a new SessionKey
            self.queryset.delete()
            sk = models.SessionKey(userkey=user_key)
            sk.save(master_key=master_key)
            key = sk.key
            self.queryset = sk

        # Encode the key using base64. (b64decode() returns a bytestring under Python 3.)
        encoded_key = base64.b64encode(key).decode()

        # Craft the response
        response = Response(
            self.serializer_class(
                self.queryset,
                context={'request': request, 'session_key': encoded_key},
            ).data,
            status=200 if preserve_key else 201,
        )

        # If token authentication is not in use, assign the session key as a cookie
        if request.auth is None:
            response.set_cookie(
                constants.SESSION_COOKIE_NAME,
                value=encoded_key,
                secure=settings.SESSION_COOKIE_SECURE,
                samesite='Strict',
                max_age=settings.LOGIN_TIMEOUT,
            )

        return response


class GenerateRSAKeyPairViewSet(ViewSet):
    """
    This endpoint can be used to generate a new RSA key pair. The keys are returned in PEM format.

        {
            "public_key": "<public key>",
            "private_key": "<private key>"
        }
    """

    serializer_class = serializers.RSAKeyPairSerializer
    permission_classes = [IsAuthenticated]

    def get_queryset(self):
        # This is only used to generate the schema
        return models.UserKey.objects.filter(user=self.request.user)

    def list(self, request):

        # Determine what size key to generate
        try:
            key_size = request.GET.get('key_size', public_key_size)
            key_size = int(key_size)
        except ValueError:
            key_size = public_key_size

        if key_size not in range(2048, 4097, 256):
            key_size = public_key_size

        # Export RSA private and public keys in PEM format
        key = RSA.generate(key_size)
        private_key = key.exportKey('PEM')
        public_key = key.publickey().exportKey('PEM')

        return Response(
            {
                'private_key': private_key,
                'public_key': public_key,
            },
        )


class GetSessionKeyViewSet(ViewSet):
    """
    Retrieve a temporary session key to use for encrypting and decrypting secrets via the API. The user's private RSA
    key is POSTed with the name `private_key`.
    This endpoint accepts one optional parameter: `preserve_key`. If True and a session key exists, the existing session
    key will be returned instead of a new one.

    Deprecation notice: This endpoint is deprecated and will be removed in a future release. Use the `SessionKeyViewSet`.
    """

    permission_classes = [IsAuthenticated]
    parser_classes = [JSONParser, FormParser, MultiPartParser]

    def create(self, request):

        # Read private key
        private_key = request.data.get('private_key', None)
        if private_key is None:
            return HttpResponseBadRequest(ERR_PRIVKEY_MISSING)

        # Validate user key
        try:
            user_key = models.UserKey.objects.get(user=request.user)
        except models.UserKey.DoesNotExist:
            return HttpResponseBadRequest(ERR_USERKEY_MISSING)
        if not user_key.is_active():
            return HttpResponseBadRequest(ERR_USERKEY_INACTIVE)

        # Validate private key
        master_key = user_key.get_master_key(private_key)
        if master_key is None:
            return HttpResponseBadRequest(ERR_PRIVKEY_INVALID)

        try:
            current_session_key = models.SessionKey.objects.get(userkey__user_id=request.user.pk)
        except models.SessionKey.DoesNotExist:
            current_session_key = None

        if current_session_key and request.data.get('preserve_key', False):

            # Retrieve the existing session key
            key = current_session_key.get_session_key(master_key)

        else:

            # Create a new SessionKey
            models.SessionKey.objects.filter(userkey__user=request.user).delete()
            sk = models.SessionKey(userkey=user_key)
            sk.save(master_key=master_key)
            key = sk.key

        # Encode the key using base64. (b64decode() returns a bytestring under Python 3.)
        encoded_key = base64.b64encode(key).decode()

        # Craft the response
        response = Response(
            {
                'session_key': encoded_key,
            },
        )

        # If token authentication is not in use, assign the session key as a cookie
        if request.auth is None:
            response.set_cookie('session_key', value=encoded_key)

        return response



# Generated by Django 4.1.5 on 2023-01-26 14:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_secrets', '0005_alter_secret_custom_field_data_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='userkey',
            name='created',
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
        migrations.AlterField(
            model_name='userkey',
            name='last_updated',
            field=models.DateTimeField(auto_now=True, null=True),
        ),
    ]



from django import forms
from netbox.forms import NetBoxModelBulkEditForm
from utilities.forms.fields import CommentField

from ..models import Secret, SecretRole

__all__ = [
    'SecretRoleBulkEditForm',
    'SecretBulkEditForm',
]


class SecretRoleBulkEditForm(NetBoxModelBulkEditForm):
    pk = forms.ModelMultipleChoiceField(queryset=SecretRole.objects.all(), widget=forms.MultipleHiddenInput)
    description = forms.CharField(max_length=200, required=False)
    comments = CommentField()

    model = SecretRole

    fieldsets = ((None, ('description',)),)

    class Meta:
        nullable_fields = ['description', 'comments']


class SecretBulkEditForm(NetBoxModelBulkEditForm):
    pk = forms.ModelMultipleChoiceField(queryset=SecretRole.objects.all(), widget=forms.MultipleHiddenInput)
    description = forms.CharField(max_length=200, required=False)
    comments = CommentField()

    model = Secret

    fieldsets = ((None, ('description',)),)

    class Meta:
        nullable_fields = ['description', 'comments']



import django_filters
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.utils.translation import gettext as _

from netbox.filtersets import NetBoxModelFilterSet
from tenancy.models import Contact
from utilities.filters import ContentTypeFilter, MultiValueCharFilter
from .constants import SECRET_ASSIGNABLE_MODELS
from .models import Secret, SecretRole

__all__ = [
    'SecretFilterSet',
    'SecretRoleFilterSet',
]

plugin_settings = settings.PLUGINS_CONFIG['netbox_secrets']


class SecretRoleFilterSet(NetBoxModelFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label='Search',
    )
    name = MultiValueCharFilter(lookup_expr='iexact')

    class Meta:
        model = SecretRole
        fields = ['id', 'name', 'slug', 'description', 'comments']

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value)
            | Q(slug__icontains=value)
            | Q(description__icontains=value)
            | Q(comments__icontains=value),
        )


if plugin_settings.get('enable_contacts', False):

    class SecretFilterSet(NetBoxModelFilterSet):
        q = django_filters.CharFilter(
            method='search',
            label='Search',
        )

        name = MultiValueCharFilter(lookup_expr='iexact')

        role_id = django_filters.ModelMultipleChoiceFilter(
            queryset=SecretRole.objects.all(),
            label='Role (ID)',
        )
        role = django_filters.ModelMultipleChoiceFilter(
            field_name='role__slug',
            queryset=SecretRole.objects.all(),
            to_field_name='slug',
            label='Role (slug)',
        )

        assigned_object_type = ContentTypeFilter()

        assigned_object_type_id = django_filters.ModelMultipleChoiceFilter(
            queryset=ContentType.objects.filter(SECRET_ASSIGNABLE_MODELS),
            label='Object type (ID)',
        )

        contact = django_filters.ModelMultipleChoiceFilter(
            field_name='contacts__contact',
            queryset=Contact.objects.all(),
            label=_('Contact'),
        )

        class Meta:
            model = Secret
            fields = [
                'id',
                'assigned_object_type',
                'assigned_object_type_id',
                'assigned_object_id',
                'role_id',
                'role',
                'name',
                'contact',
                'description',
                'comments',
                '_object_repr',
            ]

        def search(self, queryset, name, value):
            if not value.strip():
                return queryset
            return queryset.filter(
                Q(name__icontains=value)
                | Q(_object_repr__icontains=value)
                | Q(description__icontains=value)
                | Q(comments__icontains=value),
            )

else:

    class SecretFilterSet(NetBoxModelFilterSet):
        q = django_filters.CharFilter(
            method='search',
            label='Search',
        )

        name = MultiValueCharFilter(lookup_expr='iexact')

        role_id = django_filters.ModelMultipleChoiceFilter(
            queryset=SecretRole.objects.all(),
            label='Role (ID)',
        )
        role = django_filters.ModelMultipleChoiceFilter(
            field_name='role__slug',
            queryset=SecretRole.objects.all(),
            to_field_name='slug',
            label='Role (slug)',
        )

        assigned_object_type = ContentTypeFilter()

        assigned_object_type_id = django_filters.ModelMultipleChoiceFilter(
            queryset=ContentType.objects.filter(SECRET_ASSIGNABLE_MODELS),
            label='Object type (ID)',
        )

        class Meta:
            model = Secret
            fields = [
                'id',
                'assigned_object_type',
                'assigned_object_type_id',
                'assigned_object_id',
                'role_id',
                'role',
                'name',
                '_object_repr',
                'description',
                'comments',
            ]

        def search(self, queryset, name, value):
            if not value.strip():
                return queryset
            return queryset.filter(
                Q(name__icontains=value)
                | Q(_object_repr__icontains=value)
                | Q(description__icontains=value)
                | Q(comments__icontains=value),
            )



import django_tables2 as tables
from netbox.tables import NetBoxTable, columns

from .models import Secret, SecretRole

#
# Secret roles
#


class SecretRoleTable(NetBoxTable):
    name = tables.Column(linkify=True)
    secret_count = columns.LinkedCountColumn(
        viewname='plugins:netbox_secrets:secret_list',
        url_params={'role_id': 'pk'},
        verbose_name='Secrets',
    )
    comments = columns.MarkdownColumn()
    tags = columns.TagColumn(url_name='plugins:netbox_secrets:secretrole_list')

    class Meta(NetBoxTable.Meta):
        model = SecretRole
        fields = (
            'pk',
            'id',
            'name',
            'secret_count',
            'description',
            'slug',
            'comments',
            'tags',
            'created',
            'last_updated',
            'actions',
        )
        default_columns = ('id', 'name', 'secret_count', 'description', 'actions')


#
# Secrets
#


class SecretTable(NetBoxTable):
    name = tables.Column(linkify=True)
    assigned_object_type = columns.ContentTypeColumn(verbose_name='Object type')
    assigned_object = tables.Column(linkify=True, orderable=False, verbose_name='Object')
    role = tables.Column(linkify=True)
    comments = columns.MarkdownColumn()
    tags = columns.TagColumn(url_name='plugins:netbox_secrets:secret_list')

    class Meta(NetBoxTable.Meta):
        model = Secret
        fields = (
            'pk',
            'id',
            'name',
            'description',
            'assigned_object_type',
            'assigned_object',
            'role',
            'comments',
            'created',
            'last_updated',
            'tags',
        )
        default_columns = (
            'pk',
            'id',
            'name',
            'description',
            'assigned_object_type',
            'assigned_object',
            'role',
            'actions',
        )



# Generated by Django 4.1.5 on 2023-01-13 11:59

import utilities.json
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_secrets', '0004_populate_secrets'),
    ]

    operations = [
        migrations.AlterField(
            model_name='secret',
            name='custom_field_data',
            field=models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder),
        ),
        migrations.AlterField(
            model_name='secretrole',
            name='custom_field_data',
            field=models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder),
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from drf_spectacular.utils import extend_schema_field
from netbox.api.fields import ContentTypeField
from netbox.api.serializers import NetBoxModelSerializer
from netbox.constants import NESTED_SERIALIZER_PREFIX
from rest_framework import serializers
from utilities.api import get_serializer_for_model

from ..constants import SECRET_ASSIGNABLE_MODELS
from ..models import *
from .nested_serializers import *

__all__ = [
    'SecretRoleSerializer',
    'SecretSerializer',
    'SessionKeySerializer',
    'SessionKeyCreateSerializer',
    'UserKeySerializer',
    'RSAKeyPairSerializer',
]


#
# User Key
#


class UserKeySerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name='plugins-api:netbox_secrets-api:userkey-detail')
    public_key = serializers.CharField()
    private_key = serializers.CharField(
        write_only=True,
    )

    display = serializers.SerializerMethodField(read_only=True)

    is_active = serializers.BooleanField(read_only=True)

    is_filled = serializers.BooleanField(read_only=True)

    class Meta:
        model = UserKey
        fields = [
            'pk',
            'id',
            'url',
            'display',
            'public_key',
            'private_key',
            'created',
            'last_updated',
            'is_active',
            'is_filled',
        ]

    @extend_schema_field(serializers.CharField())
    def get_display(self, obj):
        return str(obj)


#
# Session Keys
#


class SessionKeySerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name='plugins-api:netbox_secrets-api:sessionkey-detail')

    display = serializers.SerializerMethodField(read_only=True)

    userkey = NestedUserKeySerializer()

    session_key = serializers.SerializerMethodField(
        read_only=True,
    )

    class Meta:
        model = SessionKey
        fields = [
            'pk',
            'id',
            'url',
            'display',
            'userkey',
            'session_key',
            'created',
        ]

    @extend_schema_field(serializers.CharField())
    def get_display(self, obj):
        return str(obj)

    @extend_schema_field(serializers.CharField())
    def get_session_key(self, obj):
        return self.context.get('session_key', None)


class SessionKeyCreateSerializer(serializers.ModelSerializer):
    private_key = serializers.CharField(
        write_only=True,
    )

    preserve_key = serializers.BooleanField(
        default=False,
        write_only=True,
    )

    class Meta:
        model = SessionKey
        fields = [
            'private_key',
            'preserve_key',
        ]


#
# Secret Roles
#


class SecretRoleSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name='plugins-api:netbox_secrets-api:secretrole-detail')
    secret_count = serializers.IntegerField(read_only=True)

    class Meta:
        model = SecretRole
        fields = [
            'id',
            'url',
            'display',
            'name',
            'slug',
            'description',
            'comments',
            'custom_fields',
            'created',
            'last_updated',
            'secret_count',
        ]


#
# Secrets
#


class SecretSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name='plugins-api:netbox_secrets-api:secret-detail')
    assigned_object_type = ContentTypeField(queryset=ContentType.objects.filter(SECRET_ASSIGNABLE_MODELS))
    assigned_object = serializers.SerializerMethodField(read_only=True)
    role = NestedSecretRoleSerializer()
    plaintext = serializers.CharField()

    class Meta:
        model = Secret
        fields = [
            'id',
            'url',
            'display',
            'assigned_object_type',
            'assigned_object_id',
            'assigned_object',
            'role',
            'name',
            'plaintext',
            'hash',
            'description',
            'comments',
            'tags',
            'custom_fields',
            'created',
            'last_updated',
        ]
        validators = []

    @extend_schema_field(serializers.DictField())
    def get_assigned_object(self, obj):
        serializer = get_serializer_for_model(obj.assigned_object, prefix=NESTED_SERIALIZER_PREFIX)
        context = {'request': self.context['request']}
        return serializer(obj.assigned_object, context=context).data

    def validate(self, data):
        # Encrypt plaintext data using the master key provided from the view context
        if data.get('plaintext'):
            s = Secret(plaintext=data['plaintext'])
            s.encrypt(self.context['master_key'])
            data['ciphertext'] = s.ciphertext
            data['hash'] = s.hash

        super().validate(data)

        return data


class RSAKeyPairSerializer(serializers.Serializer):
    public_key = serializers.CharField()
    private_key = serializers.CharField()



from django.db import migrations


def populate_secretroles(apps, schema_editor):
    """Populate the SecretRole model with data from the SecretStore model."""
    try:
        SecretRoleOld = apps.get_model('netbox_secretstore', 'SecretRole')
    except LookupError:
        # Skip if the old model doesn't exist
        return
    SecretRole = apps.get_model('netbox_secrets', 'SecretRole')

    # Retrieve the necessary data from SecretStore objects
    roles = SecretRoleOld.objects.values(
        'id',
        'name',
        'slug',
        'description',
        'created',
        'last_updated',
        'custom_field_data',
    )

    # Queue SecretRoles to be created
    roles_to_create = []
    for i, role in enumerate(roles, start=1):
        roles_to_create.append(
            SecretRole(
                id=role['id'],
                name=role['name'],
                slug=role['slug'],
                description=role['description'],
                created=role['created'],
                last_updated=role['last_updated'],
                custom_field_data=role['custom_field_data'],
            ),
        )

    # Bulk create the role objects
    SecretRole.objects.bulk_create(roles_to_create, batch_size=100)


class Migration(migrations.Migration):
    dependencies = [
        ('netbox_secrets', '0002_populate_userkeys'),
    ]

    operations = [
        migrations.RunPython(code=populate_secretroles, reverse_code=migrations.RunPython.noop),
    ]
